# include	"ICAPI.h"

# include	<string.h>

static const char IC_PREFERENCE_TABLE[] = "ic_preference";
static const char IC_KEY_FIELD[]		= "key";
static const char IC_DATA_FIELD[]		= "data";

static const char IC_URL_HELPER_TABLE[] = "ic_url_helper";
static const char IC_SCHEME_FIELD[] = "scheme";
static const char IC_HELPER_FIELD[] = "helper";


//=========================================================================
//	* ICInstance::ICInstance
//=========================================================================
//	Initialize an ICInstance. The inCreatorCode argument is used for many
//	interesting things. It should be the creator code of the application or
//	addon which is opening IC.
//
//	Create a new ICInstance in the following fashion:
//
//		ICInstance* myIC = new ICInstance( MY_CREATOR_CODE );

ICInstance::ICInstance(
	long inCreatorCode )
{
	mCreatorCode = inCreatorCode;
	InitDatabaseTables( );
}


//=========================================================================
//	* ICInstance::~ICInstance
//=========================================================================
//	Perform any needed cleanup. ICInstances should be destroyed using the
//	delete operator.

ICInstance::~ICInstance( )
{
}


//=========================================================================
//	* ICInstance::InitDatabaseTables
//=========================================================================
//	Create our database tables if they aren't already around.

ICError ICInstance::InitDatabaseTables( )
{
	BDatabase* db = boot_volume( ).Database( );
	BTable* preferencesTable;
	BTable* helperTable;

	db->Sync( );
	
		// Set up table of IC preferences
	preferencesTable = db->FindTable( IC_PREFERENCE_TABLE );
	if ( !preferencesTable )
		preferencesTable = db->CreateTable( IC_PREFERENCE_TABLE );
	if ( preferencesTable->FieldKey( IC_KEY_FIELD ) == B_ERROR )
		preferencesTable->AddStringField( IC_KEY_FIELD, B_INDEXED_FIELD );
	if ( preferencesTable->FieldKey( IC_DATA_FIELD ) == B_ERROR )
		preferencesTable->AddRawField( IC_DATA_FIELD );
	
		// Set up table of IC URL helpers
	helperTable = db->FindTable( IC_URL_HELPER_TABLE );
	if ( !helperTable )
		helperTable = db->CreateTable( IC_URL_HELPER_TABLE );
	if ( helperTable->FieldKey( IC_SCHEME_FIELD ) == B_ERROR )
		helperTable->AddStringField( IC_SCHEME_FIELD, B_INDEXED_FIELD );
	if ( helperTable->FieldKey( IC_HELPER_FIELD ) == B_ERROR )
		helperTable->AddStringField( IC_HELPER_FIELD, B_INDEXED_FIELD );
		
	return B_NO_ERROR;
}


//=========================================================================
//	* ICInstance::GetPref
//=========================================================================
//	Return the data for a preference, given the key. This version of the
//	call requires you to supply a buffer and a buffer size, and returns
//	the actual size of the pref, plus the data if it fits.
//
//	GetPrefMalloc( ) is strongly prefered, especially for variable length
//	data.

long ICInstance::GetPref(
	const char*		inKey,
	ICAttr*			outAttr,
	void*			outPrefData,
	long*			ioPrefDataSize ) const
{
	ICError err;
	void* dataPtr;
	long dataSize;
	
	err = GetPrefMalloc( inKey, outAttr, &dataPtr, &dataSize );
	if ( err == B_NO_ERROR )
	{
		if ( *ioPrefDataSize >= dataSize )
			memmove( outPrefData, dataPtr, dataSize );
		else
			err = IC_TRUCATED_DATA;
		*ioPrefDataSize = dataSize;
		free( dataPtr );
	}
	
	return err;
}			


//=========================================================================
//	* ICInstance::GetPrefMalloc
//=========================================================================
//	Return the data for a preference, given the key. This version allocates
//	a memory block using malloc to hold the preference data. Make you you
//	dispose of it with free(), not operator delete.
//
//	No memory will be allocated and outNewPrefData will equal NULL if an
//	error is returned.

long ICInstance::GetPrefMalloc(
	const char*		inKey,
	ICAttr*			outAttr,
	void**			outNewPrefData,
	long*			outPrefDataSize ) const
{
	ICError err;

	BDatabase* db = boot_volume( ).Database( );
	BTable* preferencesTable = db->FindTable( IC_PREFERENCE_TABLE );
	BQuery* query = new BQuery( );
	
	*outNewPrefData = NULL;
	
	if ( query == NULL )
		return IC_MEMORY_ERROR;
	
	query->AddTable( preferencesTable );
	query->PushField( IC_KEY_FIELD );
	query->PushString( inKey );
	query->PushOp( B_EQ );
	
	err = query->FetchOne( );
	if ( err == B_NO_ERROR && query->CountRecordIDs( ) != 1 )
		err = IC_NO_SUCH_PREF;

	if ( err == B_NO_ERROR )
	{
		BRecord* record = new BRecord( db, query->RecordIDAt( 0 ) );
		if ( record == NULL )
			err = IC_MEMORY_ERROR;
		else
		{
			err = record->Error( );	
			if ( err == B_NO_ERROR )
			{
				long size;
				void* data = record->FindRaw( IC_DATA_FIELD, outPrefDataSize );
				*outNewPrefData = malloc( *outPrefDataSize );
				if ( *outNewPrefData == NULL )
					err = IC_MEMORY_ERROR;
				else
					memmove( *outNewPrefData, data, *outPrefDataSize );
			}		
			delete record;
		}
	}
	
	if ( err != B_NO_ERROR && *outNewPrefData != NULL )
	{
		free( *outNewPrefData );
		*outNewPrefData = NULL;
	}
	delete query;
	return err;
}

long ICInstance::SetPref(
	const char*		inKey,
	ICAttr			inAttr,
	void*			inPrefData,
	long			inPrefDataSize )
{
	ICError err;

	BDatabase* db = boot_volume( ).Database( );
	BTable* preferencesTable = db->FindTable( IC_PREFERENCE_TABLE );
	BQuery* query = new BQuery( );
	
	if ( query == NULL )
		return IC_MEMORY_ERROR;
	
	query->AddTable( preferencesTable );
	query->PushField( IC_KEY_FIELD );
	query->PushString( inKey );
	query->PushOp( B_EQ );
	
	err = query->FetchOne( );
	if ( err == B_NO_ERROR && query->CountRecordIDs( ) != 1 )
		err = IC_NO_SUCH_PREF;

	if ( err == B_NO_ERROR )
	{
		BRecord* record = new BRecord( db, query->RecordIDAt( 0 ) );
		if ( record == NULL )
			err = IC_MEMORY_ERROR;
		else
		{
			err = record->Error( );	
			if ( err == B_NO_ERROR )
				record->SetRaw( IC_DATA_FIELD, inPrefData, inPrefDataSize );
			record->Commit( );		
			delete record;
		}
	}
	
	delete query;
	return err;
}

long ICInstance::DisplayURL( const char* inURL, int inFlags ) const
{
	return B_NO_ERROR;
}

typedef BRecord* BRecordPtr_;

void ICInstance::RegisterURLSchemes( const char** inSchemeList )
{
	BDatabase* db = boot_volume( ).Database( );
	BTable* helperTable = db->FindTable( IC_URL_HELPER_TABLE );
	BQuery* query = new BQuery( );
	char creatorString[sizeof(long) + 1];
	BRecord** recordList;
	
	memcpy( creatorString, &mCreatorCode, sizeof(long) );
	creatorString[sizeof(long)] = '\0';
	
		// Query for records
	query->AddTable( helperTable );
	query->PushField( IC_HELPER_FIELD );
	query->PushString( creatorString );
	query->PushOp( B_EQ );
	query->Fetch( );
	
		// Fetch records
	recordList = new BRecordPtr_[query->CountRecordIDs( )];
	for ( int j = query->CountRecordIDs( ) - 1; j >= 0; j-- )
		recordList[j] = new BRecord( db, query->RecordIDAt( j ) );
		
		// Compare scheme list with records
	for ( int i = 0; inSchemeList[i] != NULL; i++ )
	{
			// Check to see if this scheme/creator is already there
		bool alreadyInDatabase = FALSE;
		for ( int j = query->CountRecordIDs( ) - 1; j >= 0; j-- )
		{
			if ( strcmp( recordList[j]->FindString( IC_SCHEME_FIELD ),
				inSchemeList[i] ) == 0 )
			{
				alreadyInDatabase = TRUE;
				break;
			}
		}
		
			// Add any new records
		if ( !alreadyInDatabase )
		{
			BRecord* newRec = new BRecord( helperTable );
			newRec->SetString( IC_SCHEME_FIELD, inSchemeList[i] );
			newRec->SetString( IC_HELPER_FIELD, creatorString );
			newRec->Commit( );
			delete newRec;
		}			
	}
	
		// Dispose of record list
	for ( int j = query->CountRecordIDs( ) - 1; j >= 0; j-- )
		delete recordList[j];
	delete [] recordList;
}
