# ifndef __ERRORUTILITIES__
# define __ERRORUTILITIES__

# ifndef USE_LONG_ERRORS
typedef short ErrorCode;
# else
typedef long ErrorCode;
# endif


//=========================================================================
//	* Common error codes & categories of errors
//=========================================================================

enum {
	noErr = 0,
	genericUnixErr,
	nullPointerErr
};

enum /* ErrorCategory */ {
	kNonfatalErr,
	kFatalErr
};


//=========================================================================
//	* Error reporting
//=========================================================================

# define ReportError(err,what) \
	_InternalReportError(err,__FILE__,__LINE__,what,kNonfatalErr)
# define ReportFatalError(err,what) \
	_InternalReportError(err,__FILE__,__LINE__,what,kFatalErr)

# ifdef __cplusplus
extern "C" {
# endif

extern void _InternalReportError( ErrorCode inErr, const char* inFileName,
	int inLineNumber, const char* inReason, int inCategory );

# ifdef __cplusplus
}
# endif


//=========================================================================
//	* Exception-handling macros
//=========================================================================
//	There's some technical preprocessor hackery going on here. =) To use
//	these macros, include this file and set your functions up as follows:
//
//	ErrorCode myFunc( )
//	{
//		ErrorCode err = noErr;
//
//		err = doSomething( );
//		RETURN_ON_X( err );
//
//	clean_up:
//		/* tidy */
//		return err;
//	}

# define BLK(code) do{code;}while(0)

# ifdef AGGRESSIVE_DEBUG
#	define RETURN(e) \
	BLK(err=e;if(err!=noErr)ReportError(err,"Error return.");goto clean_up)
# else
#	define RETURN(e) BLK(err=e;goto clean_up)
# endif

# define RETURN_ON(cond,e)		BLK(if (cond) RETURN(e))
# define RETURN_ON_ERR(e)		BLK(ErrorCode _e=e; if (_e!=noErr) RETURN(_e))
# define RETURN_ON_NEGATIVE(i)	BLK(if (i<0) RETURN(genericUnixErr))
# define RETURN_ON_NONZERO(i)	BLK(if (i!=0) RETURN(genericUnixErr))
# define RETURN_ON_NULL(p)		BLK(if (p==NULL) RETURN(nullPointerErr))


//=========================================================================
//	* C++ Exception-handling support
//=========================================================================
//	These macros provide support for C++ exception handling to be used in
//	a fashion similar to the above macros.

# ifdef __cplusplus

class CException
{
private:
	ErrorCode	mErr;
	const char* mFilename;
	int			mLineNumber;

public:
				CException( ErrorCode inErr, const char* inFilename, int inLineNumber );
					
	void		Report( );
	ErrorCode	GetErrorCode( ) { return mErr; }
};

inline CException::CException( ErrorCode inErr, const char* inFilename, int inLineNumber )
	: mErr(inErr), mFilename(inFilename), mLineNumber(inLineNumber) { }

inline void CException::Report( )
{
# ifdef MIXED_IO
		// Flush cerr before calling code using stderr. Portable.
	cerr.sync_with_stdio( );
# endif
	_InternalReportError( mErr, mFilename, mLineNumber, "exception thrown", kNonfatalErr );
}

# define THROW(e) 				BLK(throw CException(e,__FILE__,__LINE__))

# define THROW_ON(cond,e)		BLK(if (cond) THROW(e))
# define THROW_ON_ERR(e)		BLK(ErrorCode _e=e; if (_e!=noErr) THOW(_e))
# define THROW_ON_NEGATIVE(i)	BLK(if (i<0) THROW(genericUnixErr))
# define THROW_ON_NONZERO(i)	BLK(if (i!=0) THROW(genericUnixErr))
# define THROW_ON_NULL(p)		BLK(if (p==NULL) THROW(nullPointerErr))

	// I wish I were programming in Dylan...
# define CLEANUP(code)			catch (...) {code throw;} code

# endif

# endif

