# include	<stdlib.h>
# include	<stdio.h>
# include	<sys/types.h>
# include	<netdb.h>
# include	<fcntl.h>
# include	<string.h>

# ifndef __BEOS_GUI__
#	include	<sys/socket.h>
#	include	<netinet/in.h>
#	include	<arpa/inet.h>
# else
#	include	<socket.h>
# endif

# include	"ErrorUtilities.h"
# include	"SocketUtilities.h"

# define	INITIAL_LINE_BUFFER		(64)


//=========================================================================
//	* ReadLineFromSocket
//=========================================================================
//	Read a single line from an inbound socket. "inSocket" is the socket to
//	read from. "outLinePtr" should point to the address of a pointer, which
//	will be set to indicate a heap-allocated block containing the line
//	delimited by a NULL. We remove any CRLF.
//
//		char* myBuffPtr;
//		err = ReadLineFromSocket( sk, &myBuffPtr );
//		RETURN_ON_ERR( err );
//		use regular string pointed to by myBuffPtr
//		free( myBuffPtr );
//
//	If an error occurs, myBuffPtr will be set to NULL.
//
//	It is important to be able to read arbitrary length lines since the
//	NNTP protocol places no upper bound on the distance between CRLF
//	pairs. This function tries to act as a "firewall" between the
//	grungy network API and the cleaner interior of the program.
//
//	XXX - We do not check to see if the incoming line length overflows
//	a long. This is an unlikely attack on client security and will
//	generally cause memory allocation to fail long before anything
//	nasty happens.

ErrorCode ReadLineFromSocket( int inSocket, char** outLinePtr )
{
	ErrorCode	err;
	int			unixstat;
	bool		foundEOL;
	int			currentBuffSize;
	long		bytesAccepted;
	long		bytesRead;
	
//	fd_set		select_set;
//	int			flags;
	
	long		i;
	char*		temp;
	
		// Set up our variables
	err = noErr;
	foundEOL = FALSE;
	currentBuffSize = INITIAL_LINE_BUFFER;
	bytesAccepted = 0;
	
		// Allocate a starting buffer
	*outLinePtr = (char*) malloc(currentBuffSize);
	RETURN_ON_NULL( *outLinePtr );
	
		// Set non-blocking reads
//	flags = fcntl( inSocket, F_GETFL, 0 );
//	RETURN_ON( flags == -1, genericUnixErr );
//	unixstat = fcntl( inSocket, F_SETFL, flags | O_NONBLOCK );
//	RETURN_ON_NEGATIVE( flags );
	
	while ( 1 )
	{
			// Wait for some data to arrive
//		fprintf( stderr, "Waiting for data...\n" );
//		FD_ZERO( &select_set );
//		FD_SET( inSocket, &select_set );
//		unixstat = select( getdtablesize( ), &select_set, NULL, NULL, NULL );
//		RETURN_ON_NEGATIVE( unixstat );
		
			// Peek at a chunk of data
		bytesRead = recv( inSocket, *outLinePtr + bytesAccepted,
			currentBuffSize - bytesAccepted, MSG_PEEK );
		RETURN_ON_NEGATIVE( bytesRead );
	
			// Scan for a CRLF pair
		for ( i = bytesAccepted; i < (bytesAccepted + bytesRead - 1); i++ )
		{
			if ( (*outLinePtr)[i] == '\r' && (*outLinePtr)[i+1] == '\n' )
			{
					// Accept data through LF, build string, return.
					// XXX - Are we guaranteed to read bytes found with MSG_PEEK?
				foundEOL = TRUE;
				bytesRead = recv( inSocket, *outLinePtr + bytesAccepted, i + 2 - bytesAccepted, 0 );
				RETURN_ON_NEGATIVE( bytesRead );
				(*outLinePtr)[i] = '\0';

#if defined AGGRESSIVE_DEBUG && !defined __BEOS_GUI__
				fprintf( stderr, "<- %s\n", *outLinePtr );
#endif

				RETURN( noErr );
			}
		}
		
			// No CRLF found, so accept all bytes seen so far (except for
			// the very last byte, which could be the first half of a CRLF).
			// XXX - Potential MSG_PEEK problem again
		bytesRead = recv( inSocket, *outLinePtr + bytesAccepted, (bytesRead - 1), 0 );
		RETURN_ON_NEGATIVE( bytesRead );
		bytesAccepted += bytesRead;
		
			// If our buffer is more than half full, double its size.
		if ( bytesAccepted > currentBuffSize / 2 )
		{
			temp = (char*) malloc( currentBuffSize * 2 );
			RETURN_ON_NULL( temp );
			memcpy( temp, *outLinePtr, currentBuffSize );
			free( *outLinePtr );
			*outLinePtr = temp;
			currentBuffSize *= 2;
		}
	}

clean_up:
	if ( err && *outLinePtr != NULL )
	{
		free( *outLinePtr );
		*outLinePtr = NULL;
	}
	
	return err;
}


//=========================================================================
//	* WriteLineToSocket
//=========================================================================
//	Write the string contained in inLineBuffer to the socket, delimiting it
//	with a CRLF.

ErrorCode WriteLineToSocket( int inSocket, const char* inLineBuffer )
{
	ErrorCode err;
	int unixstat;
	
	err = noErr;
	
	unixstat = send( inSocket, inLineBuffer, strlen(inLineBuffer), 0 );
	RETURN_ON_NEGATIVE( unixstat );
	unixstat = send( inSocket, "\r\n", 2, 0 );
	RETURN_ON_NEGATIVE( unixstat );

#if defined AGGRESSIVE_DEBUG && !defined __BEOS_GUI__
	fprintf( stderr, "-> %s\n", inLineBuffer );
#endif
	
clean_up:
	return err;
}


//=========================================================================
//	* ReadFinalResponseLine
//=========================================================================
//	Read a series of response lines from the socket, ignoring those which
//	are of the form "DDD-" where D is a digit. Return the first line for
//	which the three digits are followed by a space.

extern ErrorCode ReadFinalResponseLine( int inSocket, char** outLinePtr )
{
	ErrorCode err;
	bool done;
	
	done = FALSE;
	do
	{
		err = ReadLineFromSocket( inSocket, outLinePtr );
		RETURN_ON_ERR( err );
	
		if ( (*outLinePtr)[3] == '-' )
			free( *outLinePtr );
		else
			done = TRUE;
	} while ( !done );

clean_up:
	return err;
}


//=========================================================================
//	* ExtractResponseCode
//=========================================================================
//	Takes a string beginning with a three digit NNTP/FTP/etc response code
//	and extracts it in integer form.
//	Currently, no error checking is performed, although such might be
//	added at a later date. Check the return code!

ErrorCode ExtractResponseCode( const char* inServerText, int* outResponseCode )
{
	char temp[4];
	
	temp[0] = inServerText[0];
	temp[1] = inServerText[1];
	temp[2] = inServerText[2];
	temp[3] = '\0';
	
	*outResponseCode = atoi( temp );
	
	return noErr;
}


//=========================================================================
//	* IsNNTPErrorCode
//=========================================================================
//	If the response code falls with the NNTP error range, return TRUE.

int IsNNTPErrorCode( int inResponseCode )
{
	return (400 <= inResponseCode) && (inResponseCode < 600);
}


//=========================================================================
//	* ReadTextLine
//=========================================================================
//	Read a line of text from the socket. If consists of a single period,
//	discard the line and return endOfTextBlockErr. If it consists of a
//	period followed by text, remove the period.

ErrorCode ReadTextLine( int inSocket, char** outLinePtr )
{
	ErrorCode err;
	
	err = ReadLineFromSocket( inSocket, outLinePtr );
	RETURN_ON_ERR( err );
	
	if ( **outLinePtr == '.' )
	{
		if ( (*outLinePtr)[1] == '\0' )
		{
				// Dispose line and signal end of text block
			free( *outLinePtr );
			*outLinePtr = NULL;
			RETURN( endOfTextBlockErr );
		}
		else
		{
				// Remove period from line
			memmove( *outLinePtr, *outLinePtr + 1, strlen(*outLinePtr) );
		}
	}

clean_up:
	return err;
}


//=========================================================================
//	* SendNNTPCommand
//=========================================================================
//	Send a generic NNTP command and check for an error response code.

ErrorCode SendNNTPCommand( int inSocket, const char* inCommand )
{
	ErrorCode err;
	int serverResponseCode;
	char* line;
	
	err = noErr;
	line = NULL;
	
	RETURN_ON_ERR( WriteLineToSocket( inSocket, inCommand ) );
	RETURN_ON_ERR( ReadLineFromSocket( inSocket, &line ) );
	RETURN_ON_ERR( ExtractResponseCode( line, &serverResponseCode ) );
	RETURN_ON( IsNNTPErrorCode( serverResponseCode ), nntpProtocolErr );
	
clean_up:
	if ( line )
		free( line );
		
	return err;	
}
