# include	<errno.h>

# include	<CTCPStream.h>
# include	<ErrorUtilities.h>

BLocker			CTCPStream::sTCPLock( "TCP Lock" );
struct hostent*	CTCPStream::sLocalHost = NULL;
ulong			CTCPStream::sLocalIP = 0;


//=========================================================================
//	* Constructors and destructor
//=========================================================================
//	Create and destroy CTCPStream objects. The single-argument constructor
//	creates unbound, unconnected threads. The two-argument constructor
//	is only used by the Accept( ) method, and assigns an alreay-open socket
//	to a CTCPStream (which takes resposibility for deleting it). Normally,
//	you create either client or server sockets. Client sockets are created
//	as follows (be sure to handle errors):
//
//		CTCPStream clientStream( "sample client" );
//		err = clientStream.Connect( "foo.com", myTCPPortNumber );
//		// do whatever...
//	
//	Server sockets are used as follows:
//
//		CTCPStream serverStream( "sample server" );
//		err = serverStream.Bind( B_PRIMARY_INTERFACE, portNumber );
//		err = serverStream.Listen( 4 ); // create four-connection backlog
//		while ( ... )
//		{
//			CTCPStream* newStream = serverStream.Accept( );
//			if ( newStream )
//			{
//				// do whatever...
//				delete newStream;
//			}
//		}

CTCPStream::CTCPStream( const char* inName )
	: CStream( inName )
{
	mSocket = socket( AF_INET, SOCK_STREAM, 0 );
	if ( mSocket < 0 )
		mError = B_ERROR;
	mIsOpen = FALSE;
}

CTCPStream::CTCPStream( const char* inName, int inSocket )
	: CStream( inName )
{
	mSocket = inSocket;
	mIsOpen = TRUE;
}

CTCPStream::~CTCPStream( )
{
	if ( State( ) == OpenState )
		Close( );
}


//=========================================================================
//	* CTCPStream::CacheLocalHost
//=========================================================================
//	Called internally to make sure we know the address of the local host.
//	XXX - We need to copy hostent. For now, just cache the IP address.

void CTCPStream::CacheLocalHost( )
{
	if ( sLocalIP == 0 )
	{
		sTCPLock.Lock( );
		char hostname[MAXHOSTNAMELEN];
		gethostname( hostname, sizeof(hostname) );
		sLocalHost = gethostbyname( hostname );
		sLocalIP = *((u_long*) sLocalHost->h_addr);
		sTCPLock.Unlock( );
	}
}


//=========================================================================
//	* CTCPStream::IsLocalIP
//=========================================================================
//	Check to see if a given address belongs to this machine.
//	XXX - We don't handle multiple interfaces well.

bool CTCPStream::IsLocalIP( ulong inIP )
{

	CacheLocalHost( );
	return (inIP == sLocalIP);
}


//=========================================================================
//	* CTCPStream::Read
//=========================================================================
//	Read a raw block of data, using semantics roughly equivalent to those
//	of Berkeley sockets. Do not return until we have received data or lost
//	the connection. Return either the number of characters read, or a
//	negative number to represent an error.

long CTCPStream::Read( void* outBuffer, long inBufferSize )
{
	int count = recv( mSocket, (char*) outBuffer, inBufferSize, 0 );

		// XXX - Fudge around the fact that a zero byte
		// read should really be an ENOTCONN. Ick. DR7 only?
	if ( count > 0 )
		mError = 0;
	else if ( count == 0 )
		mError = ENOTCONN;
	else
		mError = errno;

	return count ? count : ENOTCONN;
}


//=========================================================================
//	* CTCPStream::Write
//=========================================================================
//	Write a block of data to a socket. Return B_NO_ERROR for success, and
//	a negative number for failure.

long CTCPStream::Write( const void* inBuffer, long inBufferSize )
{
	mError = send( mSocket, (char*) inBuffer, inBufferSize, 0 );

		// XXX - Does send return a character count? Fix it.
	if ( mError > 0 )
		mError = B_NO_ERROR;

	return mError;
}


//=========================================================================
//	* CTCPStream::Close
//=========================================================================
//	Close a stream. The connection is lost.

void CTCPStream::Close( )
{
	if ( State( ) == OpenState )
	{
		closesocket( mSocket );
		mIsOpen = FALSE;
	}
	else
		/* ERROR */;
}

long CTCPStream::State( ) const
{
	return mIsOpen ? OpenState : ClosedState;
}



//=========================================================================
//	* Functions specific to either clients or servers
//=========================================================================
//	See the example code in the comment about the constructors to see how
//	these are used. They work just like Berkeley sockets.

long CTCPStream::Connect( const char* inHostname, int inPort )
{
	long err = B_NO_ERROR;
	int unixstat;
	sTCPLock.Lock( );

		// Look up the IP address of the server
	struct hostent* server_info = gethostbyname( inHostname );
	RETURN_ON_NULL( server_info );
	ulong server_ip = *((u_long*) server_info->h_addr);

		// Check to see if we're connecting to ourselves
	if ( IsLocalIP( server_ip ) )
		server_ip = LOOPBACK_INTERFACE;
 
		// Build the server's address
	struct sockaddr_in server_addr;
	bzero( (char*) &server_addr, sizeof(struct sockaddr_in) );
	server_addr.sin_family		= AF_INET;
	server_addr.sin_port		= inPort;
	server_addr.sin_addr.s_addr	= server_ip;

		// Connect to the server
	unixstat = connect( mSocket, (sockaddr*) &server_addr,
		sizeof(struct sockaddr_in) );
	RETURN_ON_NEGATIVE( err );
	
	mIsOpen = TRUE;
	
clean_up:
	sTCPLock.Unlock( );
	mError = err;
	return mError;
}

long CTCPStream::Bind( ulong inInterface, int inPort )
{
	if ( inInterface == PRIMARY_INTERFACE )
	{
		CacheLocalHost( );
		inInterface = sLocalIP;
	}

	struct sockaddr_in our_addr;
	bzero( (char*) &our_addr, sizeof(struct sockaddr_in) );
	our_addr.sin_family = AF_INET;
	our_addr.sin_port = inPort;
	our_addr.sin_addr.s_addr = inInterface;
	
	sTCPLock.Lock( );
	mError = bind( mSocket, (struct sockaddr*) &our_addr,
		sizeof(struct sockaddr_in) );
	sTCPLock.Unlock( );
	return mError;
}

long CTCPStream::Listen( int inBacklog )
{
	sTCPLock.Lock( );
	mError = listen( mSocket, inBacklog );
	sTCPLock.Unlock( );
	return mError;
}

CTCPStream* CTCPStream::Accept( )
{
	long err = B_NO_ERROR;
	CTCPStream* newStream = NULL;

	struct sockaddr_in client_addr;
	int addrsize = sizeof(sockaddr_in);
	
	int newSocket = accept( mSocket, (struct sockaddr*) &client_addr,
		&addrsize );
	RETURN_ON_NEGATIVE( newSocket );
	
	newStream = new CTCPStream( "accepted client", newSocket );
	RETURN_ON_NULL( newStream );
	RETURN_ON_ERR( newStream->Error( ) );
	
clean_up:
	if ( err )
	{
		if ( newStream )
			delete newStream;
		return NULL;
	}

	return newStream;
}

