# include	<Debug.h>

# include	<CThread.h>


//=========================================================================
//	* CThread
//=========================================================================
//	This class implements a convenient thread object with a few
//	improvements to standard kernel thread handling. No error-checking
//	will be performed until the BeOS supports exceptions for user programs.


//=========================================================================
//	* Constructor and destructor
//=========================================================================
// To start a newly created thread running, call Resume( ) or WaitFor( ).

CThread::CThread( const char* inName, long inPriority )
	: mDeathLock( string("l>") + inName )
{
	mIsThreadFinished = FALSE;
	mThreadID = spawn_thread( &EnterThreadCallback, inName, inPriority, this );
	if ( mThreadID < B_NO_ERROR )
		/* ERROR */;
}

CThread::~CThread( )
{
	if ( !mIsThreadFinished )
		Kill( );
}


//=========================================================================
//	* Information functions
//=========================================================================
//	These functions return various pieces of information about a thread.
//	They should not be called on threads which have finished, with the
//	exception of State( ), which will report C_THREAD_FINISHED if the
//	thread has finished, and return a regular thread_state value otherwise.

void CThread::Info( thread_info& outThreadInfo )
{
	ASSERT( !mIsThreadFinished );
	if ( get_thread_info( ID( ), &outThreadInfo ) != B_NO_ERROR )
		/* ERROR */;
}

string CThread::Name( )
{
	thread_info info;
	Info( info );
	return info.name;	
}

long CThread::Priority( )
{
	thread_info info;
	Info( info );
	return info.priority;	
}

long CThread::State( sem_id* outBlockingSemaphore )
{
	if ( mIsThreadFinished )
		return C_THREAD_FINISHED;

	thread_info info;
	Info( info );
	if ( outBlockingSemaphore != NULL && info.state == B_THREAD_WAITING )
		*outBlockingSemaphore = info.sem;
	return info.state;
}


//=========================================================================
//	* Information setters
//=========================================================================
//	Set common thread parameters. Do not call on finished threads.

void CThread::SetName( const char* inName )
{
	ASSERT( !mIsThreadFinished );
	if ( rename_thread( ID( ), inName ) != B_NO_ERROR )
		/* ERROR */;
}

void CThread::SetPriority( long inPriority )
{
	ASSERT( !mIsThreadFinished );
	if ( set_thread_priority( ID( ), inPriority ) != B_NO_ERROR )
		/* ERROR */;
}


//=========================================================================
//	* CThread::Suspend and CThread::Resume
//=========================================================================

void CThread::Suspend( )
{
	ASSERT( !mIsThreadFinished );
	if ( suspend_thread( ID( ) ) != B_NO_ERROR )
		/* ERROR */;	
}

void CThread::Resume( )
{
	ASSERT( !mIsThreadFinished );
	if ( resume_thread( ID( ) ) != B_NO_ERROR )
		/* ERROR */;	
}


//=========================================================================
//	* CThread::Kill( )
//=========================================================================
//	Ensure that the thread is not running, killing it if needed. We get the
//	deathlock to insure that the thread is killed atomically. See WaitFor()
//	and EnterThread( ). We call ThreadFinished to notify the object that
//	the thread has stopped. **Note that threads may not kill themselves.**
//
//	XXX - Does not work on newly created but unrun threads and maybe other
//	suspended threads. Start a thread before you try to kill it, OK?

void CThread::Kill( )
{
	mDeathLock.Lock( );
	if ( !mIsThreadFinished )
	{
		ThreadFinished( );
		mIsThreadFinished = TRUE;		
		if ( kill_thread( ID( ) ) != B_NO_ERROR )
			/* ERROR */;
	}
	mDeathLock.Unlock( );
}


//=========================================================================
//	* CThread::WaitFor( )
//=========================================================================
//	Wait until the thread exits. If the thread has already finished, return
//	immediately. We get the deathlock to avoid race conditions with thread
//	death. See Kill( ) and EnterThread( ).

void CThread::WaitFor( )
{
	long scratch;

	mDeathLock.Lock( );
	if ( !mIsThreadFinished )
		if ( wait_for_thread( ID( ), &scratch ) != B_NO_ERROR )
			/* ERROR */;	
	mDeathLock.Unlock( );
}


//=========================================================================
//	* CThread::EnterThreadCallback and CThread::EnterThread
//=========================================================================
//	These functions are invoked by the system when the thread is run. They
//	pass control to RunThread( ), which contains a subclass' thread code,
//	and thence to ThreadFinished( ), which can be overriden to perform
//	cleanup. ThreadFinished is also called by Kill( ). We get the deathlock
//	to avoid race conditions with thread death. See Kill( ) and WaitFor( ).

long CThread::EnterThreadCallback( void* inThreadObject )
{
	((CThread*) inThreadObject)->EnterThread( );
	return B_NO_ERROR;
}

void CThread::EnterThread( )
{
	RunThread( );
	mDeathLock.Lock( );
	if ( !mIsThreadFinished )
	{
		ThreadFinished( );
		mIsThreadFinished = TRUE;
	}
	mDeathLock.Unlock( );
}
